from __future__ import with_statement

import os
import sys
import threadpool
import unittest
import logging
import shutil

# TODO:
# figure out how to use the code in test.util
# to configure the environment correctly
#APP_ROOT = os.getenv('APP_ROOT')

#POMSET_ROOT = "%s/pomsets/src" % APP_ROOT
#sys.path.insert(0, POMSET_ROOT)

if not os.getenv('HADOOP_HOME'):
    os.environ['HADOOP_HOME'] = os.path.sep.join(['', 'hadoop'])


import currypy
import pypatterns.command as CommandPatternModule

import cloudpool
import cloudpool.shell as ShellModule

import pomsets.automaton as AutomatonModule
import pypatterns.filter as FilterModule

import pomsets.command as TaskCommandModule
import pomsets.definition as DefinitionModule
import pomsets.library as DefinitionLibraryModule
import pomsets.parameter as ParameterModule
import pomsets.task as TaskModule

import pomsets.hadoop as HadoopModule

import definition as GeneratePomsetsModule
import test.TestExecute as BaseModule
# import test.util as UtilsModule


#def getHadoopExecutable():
#    return os.path.join(os.getenv('HADOOP_HOME'), 'bin', 'hadoop')


if not os.getenv('HADOOP_HOME'):
    os.environ['HADOOP_HOME'] = os.path.sep.join(['', 'hadoop'])

class TestHadoopBase(BaseModule.BaseTestClass):
    """
    Verifies that we can execute Hadoop nodes
    """
    
    def setUp(self):

        self.assertTrue(os.path.exists(HadoopModule.getExecutablePath()))

        BaseModule.BaseTestClass.setUp(self)

        # TODO:
        # will need to place test data into HDFS
        
        self.removeOutputFiles()
        return
    
    def tearDown(self):
        self.removeOutputFiles()
        BaseModule.BaseTestClass.tearDown(self)
        return
    
    def fileExists(self, file):
        f = os.system('%s fs -stat %s' % (HadoopModule.getExecutablePath(), file))
        return f is 0
    
    

    def createCommandBuilderMap(self):
        defaultCommandBuilder = TaskCommandModule.CommandBuilder(
            TaskCommandModule.buildCommandFunction_commandlineArgs
        )
        
        #hadoopJarCommandBuilder =  HadoopModule.JarCommandBuilder(
        #    buildCommandFunction=TaskCommandModule.buildCommandFunction_commandlineArgsOnly
        #)

        #hadoopPipesCommandBuilder = HadoopModule.PipesCommandBuilder(
        #    buildCommandFunction=TaskCommandModule.buildCommandFunction_commandlineArgsOnly
        #)

        commandBuilderMap = {
            'shell process':defaultCommandBuilder,
            'print task':defaultCommandBuilder,
            #'hadoop jar':hadoopJarCommandBuilder,
            #'hadoop pipes':hadoopPipesCommandBuilder
        }
        return commandBuilderMap
    
    

    def removeOutputFiles(self):
        return
    
    # END class TestHadoopBase
    pass


class TestHadoop1(TestHadoopBase, unittest.TestCase):
    """
    Verifies that we can execute Hadoop nodes
    """
    
    DIR_OUTPUT = 'TestHadoop1-output'

    def setUp(self):
        TestHadoopBase.setUp(self)
        return

    def tearDown(self):
        TestHadoopBase.tearDown(self)
        return


    def createTask(self, definition):
        task = TaskModule.AtomicTask()
        task.definition(definition)
        task.setParameterBinding('input file', ['input'])
        task.setParameterBinding('output file', [TestHadoop1.DIR_OUTPUT])
        return task
    
    
    # setup hadoop
    # create the pomset containing hadoop nodes
    # execute the pomset
    def createDefinition(self):
        definition = GeneratePomsetsModule.createHadoopWordcountDefinition()
        return definition
    
    def getPicklePath(self):
        return os.path.sep + \
               os.path.join('tmp', 'TestExecute.TestHadoop1.testExecute2')
    
    def assertPostExecute(self):
        BaseModule.BaseTestClass.assertPostExecute(self)
        
        # now verify that the output file exists
        self.assertTrue(self.fileExists(TestHadoop1.DIR_OUTPUT))
        self.assertTrue(
            self.fileExists(
                os.path.join(TestHadoop1.DIR_OUTPUT, 'part-r-00000')))
        
        # TODO:
        # copy the resulting output
        # and verify the results
        
        return
    
    def removeOutputFiles(self):
        # ensure that the outdir does not exist
        if self.fileExists(TestHadoop1.DIR_OUTPUT):
            os.system('%s fs -rmr %s' % (HadoopModule.getExecutablePath(), TestHadoop1.DIR_OUTPUT))
        return
    
    # END class TestHadoop1
    pass


class TestHadoopStreaming1(TestHadoopBase, unittest.TestCase):
    """
    Verifies that we can execute Hadoop streaming nodes
    """

    DIR_OUTPUT = 'TestHadoopStreaming1-output'
    
    def setUp(self):
        TestHadoopBase.setUp(self)
        return

    def tearDown(self):
        TestHadoopBase.tearDown(self)
        return

    def createTask(self, definition):
        task = TaskModule.AtomicTask()
        task.definition(definition)
        task.setParameterBinding('input file', ['input'])
        task.setParameterBinding('output file', [TestHadoopStreaming1.DIR_OUTPUT])
        task.setParameterBinding(
            'reducer', 
            [os.path.join('resources', 'testdata', 'TestHadoop', 'wordcount', 'reducer.py')])
        task.setParameterBinding(
            'mapper', 
            [os.path.join('resources', 'testdata', 'TestHadoop', 'wordcount', 'mapper.py')])
        
        return task
    
    
    # setup hadoop
    # create the pomset containing hadoop nodes
    # execute the pomset
    def createDefinition(self):
        definition = GeneratePomsetsModule.createHadoopStreamingDefinition()
        return definition
    
    def getPicklePath(self):
        return os.path.sep + \
               os.path.join('tmp', 'TestExecute.TestHadoopStreaming1.testExecute2')
    
    def assertPostExecute(self):
        BaseModule.BaseTestClass.assertPostExecute(self)
        
        # now verify that the output file exists
        self.assertTrue(self.fileExists(TestHadoopStreaming1.DIR_OUTPUT))
        self.assertTrue(
            self.fileExists(
                os.path.join(TestHadoopStreaming1.DIR_OUTPUT, 'part-00000')))
        
        # TODO:
        # copy the resulting output
        # and verify the results
        
        return
    
    def removeOutputFiles(self):
        # ensure that the outdir does not exist
        if self.fileExists(TestHadoopStreaming1.DIR_OUTPUT):
            os.system('%s fs -rmr %s' % (HadoopModule.getExecutablePath(), TestHadoopStreaming1.DIR_OUTPUT))
        return
    
    # END class TestHadoopStreaming1
    pass


def main():
    
    # UtilsModule.configLogging()

    suite = unittest.TestSuite()

    #suite.addTest(unittest.makeSuite(TestHadoop1, 'test'))
    suite.addTest(unittest.makeSuite(TestHadoopStreaming1, 'test'))
    #suite.addTest(unittest.makeSuite(TestHadoopPipes1, 'test'))
    
    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()




from __future__ import with_statement

import os
import sys
import threadpool
import unittest
import logging
import shutil

import currypy
import pypatterns.command as CommandPatternModule

import cloudpool as CloudModule
import cloudpool.shell as ShellModule

import pomsets.automaton as AutomatonModule
import pypatterns.filter as FilterModule

import pomsets.command as TaskCommandModule
import pomsets.definition as DefinitionModule
import pomsets.error as ErrorModule
import pomsets.library as DefinitionLibraryModule
import pomsets.parameter as ParameterModule
import pomsets.task as TaskModule

import pomsets.test_utils as GeneratePomsetsModule




def createEchoDefinition():
    command = ['/bin/echo']
    executable = TaskCommandModule.Executable()
    executable.stageable(False)
    # executable.command(command)
    # executable.path(command[0])
    executable.path(command)
    executable.staticArgs([])
    
    definition = DefinitionModule.createShellProcessDefinition(
        inputParameters = {
            'item to echo':{ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True}
            },
        parameterOrderings = None,
        executable = executable
    )
    definition.name('echo')
    return definition

DEFINITION_ECHO = createEchoDefinition()

DEFINITION_WORDCOUNT = GeneratePomsetsModule.createWordCountDefinition()

DEFINITION_WORDCOUNT_REDUCE = GeneratePomsetsModule.createWordCountReduceDefinition()




class BaseTestClass(object):

    def setUp(self):
        automaton = AutomatonModule.Automaton()
        automaton.setThreadPool(
            None, 
            CloudModule.Pool(self.getNumWorkersToInstantiate()))
        automaton.commandManager(CommandPatternModule.CommandManager())
        self.automaton = automaton
        return

    def tearDown(self):
        return

    def getNumWorkersToInstantiate(self):
        return 1

    def createCommandBuilderMap(self):
        commandBuilder = TaskCommandModule.CommandBuilder(
            TaskCommandModule.buildCommandFunction_commandlineArgs
        )
        commandBuilderMap = {
            'shell process':commandBuilder,
            'print task':commandBuilder
        }
        return commandBuilderMap

    def createExecuteEnvironment(self):
        return ShellModule.LocalShell()

    
    def getRequestContext(self, task=None, executeEnvironment=None):
        
        automaton = self.automaton
        commandBuilderMap = self.createCommandBuilderMap()

        if executeEnvironment is None:
            executeEnvironment = self.createExecuteEnvironment()

        requestContext = {
            'task':task,
            'command builder map':commandBuilderMap,
            'execute environment':executeEnvironment
        }
        
        return requestContext
    
    
    def executeTask(self, task, 
                    executeEnvironment=None):
            
        automaton = self.automaton
        successCallback = automaton.getPostExecuteCallbackFor(task)
        errorCallback = automaton.getErrorCallbackFor(task)
        executeTaskFunction = automaton.getExecuteTaskFunction(task)
        
        requestContext = self.getRequestContext(
            task=task, executeEnvironment=executeEnvironment
        )

        request = threadpool.WorkRequest(
            executeTaskFunction,
            args = [],
            kwds = requestContext,
            callback = successCallback,
            exc_callback = errorCallback
        )

        threadPool = automaton.getThreadPoolUsingRequest(request)
        request.kwds['thread pool'] = threadPool
        
        task.workRequest(request)

        task.automaton(self.automaton)

        threadPool.putRequest(request)
        threadPool.wait()

        return request

    def assertPreExecute(self):
        return

    def assertPostExecute(self):
        task = self.request.kwds['task']
        if isinstance(task, TaskModule.CompositeTask):
            childTasks = [x for x in task.getErroredChildTasks()]
            assert not self.request.exception, childTasks[0].workRequest().kwds['exception stack trace']
        else:
            assert not self.request.exception

        return

    def testExecute1(self):

        self.assertPreExecute()

        definition = self.createDefinition()
        task = self.createTask(definition)
        request = self.executeTask(task)

        self.request = request

        self.assertPostExecute()

        return 


    def testExecute2(self):

        self.assertPreExecute()

        definition = self.createDefinition()

        definition = GeneratePomsetsModule.pickleAndReloadDefinition(
            self.getPicklePath(),
            definition
        )

        task = self.createTask(definition)
        request = self.executeTask(task)
        self.request = request

        self.assertPostExecute()
        return 

    
    # END class BaseTestClass
    pass


class TestCase1(BaseTestClass, unittest.TestCase):
    """
    execute of atomic function
    """

    def createDefinition(self):
        return DEFINITION_ECHO

    def createTask(self, definition):
        task = TaskModule.AtomicTask()
        task.definition(definition)
        task.setParameterBinding('item to echo', ['foo'])
        return task


    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase1.testExecute2')

    # END TestCase1
    pass


class TestCase2(BaseTestClass, unittest.TestCase):
    """
    execute of atomic function
    """

    def createDefinition(self):
        definition = DEFINITION_ECHO
        return definition

    def createTask(self, definition):
        task = TaskModule.AtomicTask()
        task.definition(definition)
        task.setParameterBinding('item to echo', ['"echoed testExecuteAtomicFunction2"'])
        return task

    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase2.testExecute2')

    # END class TestCase2
    


class TestCase4(BaseTestClass, unittest.TestCase):
    """
    execute of composite function
    """

    def createDefinition(self):
        atomicDefinition = DEFINITION_ECHO

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        parameter = ParameterModule.DataParameter(
            id='item to echo', 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        compositeDefinition.addParameter(parameter)
        
        node = compositeDefinition.createNode(id='node')
        node.definitionToReference(atomicDefinition)
        node.name('echo')

        compositeDefinition.connectParameters(
            compositeDefinition, 'item to echo',
            node, 'item to echo'
        )
        return compositeDefinition

    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding('item to echo', ['foo'])

        return compositeTask


    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase4.testExecute2')


    # END class TestCase4
    pass



    

class TestCase8(BaseTestClass, unittest.TestCase):
    """
    execute of composite function
    """

    def createDefinition(self):
        atomicDefinition = DEFINITION_ECHO

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        parameter = ParameterModule.DataParameter(
            id='item to echo', 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        compositeDefinition.addParameter(parameter)
        
        for index in range(3):
            node = compositeDefinition.createNode(id='node%s' % index)
            node.definitionToReference(atomicDefinition)
            compositeDefinition.connectParameters(
                compositeDefinition, 'item to echo',
                node, 'item to echo'
            )
            node.name('echo %s' % index)

        return compositeDefinition

    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding('item to echo', ['foo'])
        
        return compositeTask


    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase8.testExecute2')

    # END class TestCase8
    pass


class TestCase9(BaseTestClass, unittest.TestCase):
    """
    execute of composite function
    """

    def createDefinition(self):
        atomicDefinition = DEFINITION_ECHO

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        parameter = ParameterModule.DataParameter(
            id='item to echo', 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        compositeDefinition.addParameter(parameter)

        nodes = []
        for index in range(3):
            node = compositeDefinition.createNode(id='node%s' % index)
            node.definitionToReference(atomicDefinition)
            compositeDefinition.connectParameters(
                compositeDefinition, 'item to echo',
                node, 'item to echo'
            )
            node.name('echo %s' % index)
            nodes.append(node)
            pass

        for sourceNode, targetNode in zip(nodes[:-1], nodes[1:]):
            compositeDefinition.connectParameters(
                sourceNode, 'temporal output',
                targetNode, 'temporal input'
            )
            pass
        return compositeDefinition

    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        
        compositeTask.setParameterBinding('item to echo', ['foo'])
        
        return compositeTask


    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase9.testExecute2')

    
    # END class TestCase9
    pass

    
class TestCase10(BaseTestClass, unittest.TestCase):
    """
    execution fails due to incomplete parameter binding 
    """

    def createDefinition(self):
        atomicDefinition = DEFINITION_ECHO

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        parameter = ParameterModule.DataParameter(
            id='item to echo', 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        compositeDefinition.addParameter(parameter)
        
        node = compositeDefinition.createNode(id='node')
        node.definitionToReference(atomicDefinition)
        node.name('echo')

        return compositeDefinition


    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding('item to echo', ['foo'])
        
        return compositeTask


    def assertPostExecute(self):
        request = self.request
        assert request.exception
        return

    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestCase10.testExecute2')

    # END class TestCase
    pass
    


class TestParameterSweep1(BaseTestClass, unittest.TestCase):
    

    def createDefinition(self):

        atomicDefinition = DEFINITION_ECHO

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        node = compositeDefinition.createNode(id='node')
        node.definitionToReference(atomicDefinition)
        node.name('echo')
        
        parameter = ParameterModule.DataParameter(
            id='item to echo', 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        parameter.setAttribute(ParameterModule.PORT_ATTRIBUTE_PARAMETERSWEEP, True)
        compositeDefinition.addParameter(parameter)
        node.isParameterSweep('item to echo', True)
        
        compositeDefinition.connectParameters(
            compositeDefinition, 'item to echo',
            node, 'item to echo'
        )

        return compositeDefinition

    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding(
            'item to echo', 
            ['"echoed testExecuteParameterSweep1 : 1"',
             '"echoed testExecuteParameterSweep1 : 2"'])
        return compositeTask
    

    def getPicklePath(self):
        return os.path.sep + \
               os.path.join('tmp', 'TestExecute.TestParameterSweep1.testExecute2')


    # END class TestParameterSweep1
    pass

    
class TestParameterSweep2(BaseTestClass, unittest.TestCase):

    BASE_DIR = os.path.join(os.getcwd(), 'resources', 'testdata', 'TestExecute')
    OUTPUT_DIR = os.path.sep + 'tmp'
    
    INPUT_FILES = ['text1', 'text2']
    OUTPUT_FILES = ['count1', 'count2']
    
    def getPicklePath(self):
        return os.path.sep + \
               os.path.join('tmp', 'TestExecute.TestParameterSweep2.testExecute2')

    def removeFile(self, file):
        try:
            os.remove(file)
        except OSError:
            pass
        return
    
    def fileExists(self, file):
        return os.path.exists(file)
    
    def setUp(self):
        BaseTestClass.setUp(self)

        # ensure that the script and the text files exist
        # remove the $BASE_DIR/count* files
        self.inputFiles = [
            os.path.join(x, y)
            for x,y in zip([TestParameterSweep2.BASE_DIR]*len(TestParameterSweep2.INPUT_FILES), 
                           TestParameterSweep2.INPUT_FILES)
        ]
        self.outputFiles = [
            os.path.join(x, y)
            for x,y in zip([TestParameterSweep2.OUTPUT_DIR]*len(TestParameterSweep2.OUTPUT_FILES), 
                           TestParameterSweep2.OUTPUT_FILES)
        ]
        
        for outputFile in self.outputFiles:
            self.removeFile(outputFile)
            pass
        
        return
    

    def tearDown(self):

        BaseTestClass.tearDown(self)
        
        self.removeFiles()
        return

    
    def removeFiles(self):
        
        # remove the $BASE_DIR/count* files
        for outputFile in self.outputFiles:
            self.removeFile(outputFile)
            pass
        return

    

    def createDefinition(self):
        atomicDefinition = DEFINITION_WORDCOUNT

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        node = compositeDefinition.createNode(id='node')
        node.definitionToReference(atomicDefinition)
        node.name('mapper')

        # TODO:
        # the info to expose the parameters on the composite definition
        # should be retrieved from the definition itself
        parameterGroup = {
            'input file':{ParameterModule.PORT_ATTRIBUTE_PARAMETERSWEEP:True}, 
            'output file':{
                ParameterModule.PORT_ATTRIBUTE_PARAMETERSWEEP:True,
                ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT:True
            }
        }
        for parameterId, attributes in parameterGroup.iteritems():
            parameter = ParameterModule.DataParameter(
                id=parameterId, 
                portDirection=ParameterModule.PORT_DIRECTION_INPUT)
            for attributeKey, attributeValue in attributes.iteritems():
                parameter.setAttribute(attributeKey, attributeValue)
                pass
            
            compositeDefinition.addParameter(parameter)
            node.isParameterSweep(parameterId, True)
        
            compositeDefinition.connectParameters(
                compositeDefinition, parameterId,
                node, parameterId
            )
            pass
        node.addParameterSweepGroup(parameterGroup.keys())

        return compositeDefinition

        
    def createTask(self, definition):

        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding(
            'input file',
            self.inputFiles
        )
        compositeTask.setParameterBinding(
            'output file',
            self.outputFiles
        )
        
        return compositeTask

        
    def assertPreExecute(self):
        BaseTestClass.assertPreExecute(self)
        
        # assert that the files do not exist
        for inputFile in self.inputFiles:
            assert self.fileExists(inputFile), 'expected inputFile %s to exist' % inputFile
            pass
        
        for outputFile in self.outputFiles:
            assert not self.fileExists(outputFile), 'expected outputFile %s to not exist' % outputFile
        
        pass


    def assertPostExecute(self):
        BaseTestClass.assertPostExecute(self)

        # assert that the files exist
        for outputFile in self.outputFiles:
            assert self.fileExists(outputFile), 'expected output file %s to exist' % outputFile
            pass
        
        return 

    
    # END class TestParameterSweep2
    pass


class TestParameterSweep3(BaseTestClass, unittest.TestCase):
    
    BASE_DIR = os.path.join(os.getcwd(), 'resources', 'testdata', 'TestExecute')
    OUTPUT_DIR = os.path.sep + 'tmp'
    
    INPUT_FILES = ['count1', 'count2']
    OUTPUT_FILES = ['count_reduce']
    
    def removeFile(self, file):
        try:
            os.remove(file)
        except OSError:
            pass
        return
    
    def fileExists(self, file):
        return os.path.exists(file)
    
    
    def setUp(self):
        BaseTestClass.setUp(self)

        # ensure that the script and the text files exist
        # remove the $BASE_DIR/count* files
        self.inputFiles = [
            os.path.join(x, y) 
            for x,y in zip([TestParameterSweep3.BASE_DIR]*len(TestParameterSweep3.INPUT_FILES), 
                           TestParameterSweep3.INPUT_FILES)
        ]
        self.outputFiles = [
            os.path.join(x, y)
            for x,y in zip([TestParameterSweep3.OUTPUT_DIR]*len(TestParameterSweep3.OUTPUT_FILES), 
                           TestParameterSweep3.OUTPUT_FILES)
        ]
        
        for outputFile in self.outputFiles:
            self.removeFile(outputFile)
            pass
        
        return
    
    def tearDown(self):
        
        BaseTestClass.tearDown(self)
        self.removeFiles()
        return
    
    def removeFiles(self):
        # remove the $BASE_DIR/count* files
        for outputFile in self.outputFiles:
            self.removeFile(outputFile)
            pass
        
        return
    


    def createDefinition(self):
            
        atomicDefinition = DEFINITION_WORDCOUNT_REDUCE

        compositeDefinition = DefinitionModule.getNewNestDefinition()
        node = compositeDefinition.createNode(id='node')
        node.definitionToReference(atomicDefinition)
        node.name('reducer')

        parameterIds = ['input files', 'output file']
        for parameterId  in parameterIds:
            parameter = ParameterModule.DataParameter(
                id=parameterId, 
                portDirection=ParameterModule.PORT_DIRECTION_INPUT)
            compositeDefinition.addParameter(parameter)
        
            compositeDefinition.connectParameters(
                compositeDefinition, parameterId,
                node, parameterId
            )
            pass

        return compositeDefinition


    def createTask(self, definition):

        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding(
            'input files',
            self.inputFiles
        )
        compositeTask.setParameterBinding(
            'output file',
            self.outputFiles
        )
        return compositeTask
        

    def assertPreExecute(self):

        BaseTestClass.assertPreExecute(self)

        for inputFile in self.inputFiles:
            assert self.fileExists(inputFile), 'expected inputFile %s to exist' % inputFile
            pass
        
        
        for outputFile in self.outputFiles:
            assert not self.fileExists(outputFile), 'expected outputFile %s to not exist' % outputFile
            
            
        return


    def assertPostExecute(self):
        BaseTestClass.assertPostExecute(self)

        # assert that the files exist
        for outputFile in self.outputFiles:
            assert self.fileExists(outputFile), 'expected output file %s to exist' % outputFile
            pass
        
        return 

    def getPicklePath(self):
        return os.path.sep + \
               os.path.join('tmp', 'TestExecute.TestParameterSweep3.testExecute2')

    # END class TestParameterSweep3
    pass


class TestParameterSweep4(BaseTestClass, unittest.TestCase):
    """
    tests combining a mapper with a reducer
    """
    
    BASE_DIR = os.path.join(os.getcwd(), 'resources', 'testdata', 'TestExecute')
    TEST_DIR = os.path.sep + 'tmp'
    
    INPUT_FILES = ['text1', 'text2']
    INTERMEDIATE_FILES = ['count1', 'count2']
    OUTPUT_FILES = ['count_reduce']

    def removeFile(self, file):
        try:
            os.remove(file)
        except OSError:
            pass
        return
    
    def fileExists(self, file):
        return os.path.exists(file)
    
    
    def setUp(self):

        BaseTestClass.setUp(self)

        # ensure that the script and the text files exist
        # remove the $BASE_DIR/count* files
        self.inputFiles = [
            os.path.join(x, y)
            for x,y in zip([TestParameterSweep4.BASE_DIR]*len(TestParameterSweep4.INPUT_FILES), 
                           TestParameterSweep4.INPUT_FILES)
        ]
        self.intermediateFiles = [
            os.path.join(x, y)
            for x,y in zip([TestParameterSweep4.TEST_DIR]*len(TestParameterSweep4.INTERMEDIATE_FILES), 
                           TestParameterSweep4.INTERMEDIATE_FILES)
        ]
        self.outputFiles = [
            os.path.join(x, y)
            for x,y in zip([TestParameterSweep4.TEST_DIR]*len(TestParameterSweep4.OUTPUT_FILES), 
                           TestParameterSweep4.OUTPUT_FILES)
        ]
        
        for outputFile in self.outputFiles+self.intermediateFiles:
            self.removeFile(outputFile)
            pass
        
        return
    
    def tearDown(self):

        BaseTestClass.tearDown(self)
        self.removeFiles()
        return
    
    def removeFiles(self):
        # remove the $BASE_DIR/count* files
        for outputFile in self.outputFiles+self.intermediateFiles:
            self.removeFile(outputFile)
            pass
        
        return
    

    def createDefinition(self):
    
        compositeDefinition = DefinitionModule.getNewNestDefinition()

        # setup the reference definition for parameter sweep
        mapperNode = compositeDefinition.createNode(id='mapper')
        mapperNode.definitionToReference(DEFINITION_WORDCOUNT)
        mapperNode.isParameterSweep('input file', True)
        mapperNode.isParameterSweep('output file', True)
        mapperNode.addParameterSweepGroup(['input file', 'output file'])
        mapperNode.name('mapper')

        reducerNode = compositeDefinition.createNode(id='reducer')
        reducerNode.definitionToReference(DEFINITION_WORDCOUNT_REDUCE)
        reducerNode.name('reducer')

        inputParameter = \
            ParameterModule.DataParameter(
                id='input file', 
                portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        inputParameter.setAttribute(
            ParameterModule.PORT_ATTRIBUTE_PARAMETERSWEEP, True)
        compositeDefinition.addParameter(inputParameter)
        compositeDefinition.connectParameters(
            compositeDefinition, 'input file',
            mapperNode, 'input file'
        )
        
        outputParameter = \
            ParameterModule.DataParameter(
                id='output file', 
                portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        compositeDefinition.addParameter(outputParameter)
        compositeDefinition.connectParameters(
            compositeDefinition, 'output file',
            reducerNode, 'output file'
        )
        
        blackboardParameter = \
            ParameterModule.BlackboardParameter('intermediate file')
        compositeDefinition.addParameter(blackboardParameter)
        compositeDefinition.connectParameters(
            compositeDefinition, 'intermediate file',
            mapperNode, 'output file'
        )
        compositeDefinition.connectParameters(
            compositeDefinition, 'intermediate file',
            reducerNode, 'input files',
        )
        
        compositeDefinition.connectParameters(
            mapperNode, 'temporal output',
            reducerNode, 'temporal input',
        )
        
        return compositeDefinition


    def createTask(self, definition):

        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        compositeTask.setParameterBinding(
            'input file',
            self.inputFiles
        )
        compositeTask.setParameterBinding(
            'output file',
            self.outputFiles
        )
        compositeTask.setParameterBinding(
            'intermediate file',
            self.intermediateFiles
        )
        return compositeTask


    def assertPostExecute(self):
        BaseTestClass.assertPostExecute(self)
        
        # assert that the files exist
        for outputFile in self.outputFiles+self.intermediateFiles:
            assert self.fileExists(outputFile), 'expected output file %s to exist' % outputFile
            pass
        
        return
        
    def getPicklePath(self):
        return os.path.sep + \
               os.path.join('tmp', 'TestExecute.TestParameterSweep4.testExecute2')

    # END class TestParameterSweep4
    pass


class TestLoop1(BaseTestClass, unittest.TestCase):
    
    BASE_DIR = os.path.join(os.getcwd(), 'resources', 'testdata', 'TestExecute')

    FILE_PREFIX = os.path.sep + os.path.join('tmp', 'loop')
    INDEX_INITIAL = 0
    INDEX_FINAL = 5
    FILE_INITIAL = '%s%s' % (FILE_PREFIX, INDEX_INITIAL)
    FILE_FINAL = '%s%s' % (FILE_PREFIX, INDEX_FINAL)
    
    def fileExists(self, file):
        return os.path.exists(file)
    
    def removeFile(self, file):
        try:
            os.remove(file)
        except OSError:
            pass
        return
    
    def setUp(self):

        BaseTestClass.setUp(self)
        
        # ensure that INITIAL_FILE exists
        shutil.copyfile(os.path.join(TestLoop1.BASE_DIR, 'text1'), 
                        TestLoop1.FILE_INITIAL)
        
        return

    
    def tearDown(self):

        self.removeFiles()
        BaseTestClass.tearDown(self)
        return
      
    def removeFiles(self):
        # ensure that INITIAL_FILE no longer exists
        # ensure that created files no longer exist
        for index in range(TestLoop1.INDEX_INITIAL, TestLoop1.INDEX_FINAL+1):
            self.removeFile('%s%s' % (TestLoop1.FILE_PREFIX, index))
            pass
        return
    
    def createDefinition(self):
        definition = GeneratePomsetsModule.createPomsetContainingLoopDefinition()
        
        # GeneratePomsetsModule.bindLoopDefinitionParameters(definition)
        nodes = [x for x in definition.nodes() if x.id() == 'loop']
        loopNode = nodes[0]
    
        # set the parameter bindings on loopNode
        loopNode.setParameterBinding(
            DefinitionModule.LoopDefinition.PARAMETER_INITIAL_STATE, 
            TestLoop1.INDEX_INITIAL)
        loopNode.setParameterBinding(
            DefinitionModule.LoopDefinition.PARAMETER_CONTINUE_CONDITION,
            "lambda x: x<%s" % TestLoop1.INDEX_FINAL)
        loopNode.setParameterBinding(
            DefinitionModule.LoopDefinition.PARAMETER_STATE_TRANSITION,
            "lambda x: x+1")
        loopNode.setParameterBinding(
            DefinitionModule.LoopDefinition.PARAMETER_STATE_CONFIGURATION,
            # set the value of input file
            # set the value of output file
            [
                # for some reason
                # exec is missing
                # we we need to do this using multiple strings
                "childTask.setParameterBinding('input file', ['/tmp/loop' + str(parentTask.getParameterBinding(DefinitionModule.LoopDefinition.PARAMETER_STATE))])",
                "childTask.setParameterBinding('output file', ['/tmp/loop' + str(parentTask.getParameterBinding(DefinitionModule.LoopDefinition.PARAMETER_STATE)+1)])"
            ]
        )
        
        return definition
    
    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        return compositeTask

    
    def getPicklePath(self):
        return os.path.sep + os.path.join('tmp', 'TestExecute.TestLoop1.testExecute2')
    

    def assertPostExecute(self):
        BaseTestClass.assertPostExecute(self)
        
        for index in range(TestLoop1.INDEX_INITIAL+1,
                           TestLoop1.INDEX_FINAL+1):
            assert self.fileExists('%s%s' % (TestLoop1.FILE_PREFIX, index))
        
        return
    
    
    # END class TestLoop1
    pass


class TestBranch1(BaseTestClass, unittest.TestCase):
    
    BASE_DIR = os.path.join(os.getcwd(), 'resources', 'testdata', 'TestExecute')
    
    FILE_PREFIX = os.path.sep + os.path.join('tmp', 'branch')
    INDEX_INITIAL = 0
    INDEX_FINAL = 1
    FILE_INITIAL = '%s%s' % (FILE_PREFIX, INDEX_INITIAL)
    FILE_FINAL = '%s%s' % (FILE_PREFIX, INDEX_FINAL)
    
    def fileExists(self, file):
        return os.path.exists(file)
    
    def removeFile(self, file):
        try:
            os.remove(file)
        except OSError:
            pass
        return
    
    def setUp(self):

        BaseTestClass.setUp(self)
        
        # ensure that INITIAL_FILE exists
        shutil.copyfile(os.path.join(TestBranch1.BASE_DIR, 'text1'), 
                        TestBranch1.FILE_INITIAL)
        
        self.assertTrue(self.fileExists(TestBranch1.FILE_INITIAL))
        return

    
    def tearDown(self):

        self.removeFiles()
        BaseTestClass.tearDown(self)
        return

    def removeFiles(self):
        # ensure that INITIAL_FILE no longer exists
        # ensure that created files no longer exist
        for index in range(TestLoop1.INDEX_INITIAL, TestBranch1.INDEX_FINAL+1):
            self.removeFile('%s%s' % (TestBranch1.FILE_PREFIX, index))
            pass
        return
    
    def createDefinition(self):
        definition = GeneratePomsetsModule.createPomsetContainingBranchDefinition()
        
        nodes = [x for x in definition.nodes() if x.id() == 'branch']
        branchNode = nodes[0]
    
        # set the parameter bindings on branchNode
        branchNode.setParameterBinding(
            DefinitionModule.BranchDefinition.PARAMETER_CONDITION_STATE, 
            0)

        branchNode.setParameterBinding(
            DefinitionModule.BranchDefinition.PARAMETER_CONDITION_FUNCTION, 
            "lambda x: x"
        )
        
        branchNode.setParameterBinding(
            DefinitionModule.BranchDefinition.PARAMETER_CONDITION_MAP, 
            "[(0, 'wordcount'), (1,'reduce')]"
        )

        
        # now bind the definitions in the branches
        wcNode = [x for x in branchNode.definitionToReference().nodes() 
                  if x.id()=='wordcount'][0]
        wcNode.setParameterBinding(
            'input file', [TestBranch1.FILE_INITIAL])
        wcNode.setParameterBinding(
            'output file', [TestBranch1.FILE_FINAL])
        
        reduceNode = [x for x in branchNode.definitionToReference().nodes() 
                      if x.id()=='reduce'][0]
        reduceNode.setParameterBinding(
            'input files', [TestBranch1.FILE_INITIAL])
        reduceNode.setParameterBinding(
            'output file', [TestBranch1.FILE_FINAL])
        
        return definition

    
    def createTask(self, definition):
        compositeTask = TaskModule.CompositeTask()
        compositeTask.definition(definition)
        taskGenerator = TaskModule.NestTaskGenerator()
        compositeTask.taskGenerator(taskGenerator)
        return compositeTask

    
    def getPicklePath(self):
        return os.path.sep + \
               os.path.join('tmp', 'TestExecute.TestBranch1.testExecute2')


    def assertPostExecute(self):
        BaseTestClass.assertPostExecute(self)
        
        
        return
    
    
    # END class TestBranch1
    pass


class TestCase11(TestCase1):
    """
    execute of atomic function
    """

    def getNumWorkersToInstantiate(self):
        return 0

    
    def executeTask(self, task, 
                    executeEnvironment=None):

        automaton = self.automaton
        successCallback = automaton.getPostExecuteCallbackFor(task)
        errorCallback = automaton.getErrorCallbackFor(task)
        executeTaskFunction = automaton.getExecuteTaskFunction(task)
        
        requestContext = self.getRequestContext(
            task=task, executeEnvironment=executeEnvironment
        )
        request = threadpool.WorkRequest(
            executeTaskFunction,
            args = [],
            kwds = requestContext,
            callback = successCallback,
            exc_callback = errorCallback
        )

        try:
            automaton.enqueueRequest(request)
            assert False
        except ErrorModule.ExecutionError, e:
            self.assertEquals('need to start thread before execution',
                              str(e))

        return request



    # END TestCase11
    pass



def main():
    
    util.configLogging()

    suite = unittest.TestSuite()

    suite.addTest(unittest.makeSuite(TestCase1, 'test'))
    suite.addTest(unittest.makeSuite(TestCase2, 'test'))
    suite.addTest(unittest.makeSuite(TestCase4, 'test'))
    suite.addTest(unittest.makeSuite(TestCase8, 'test'))
    suite.addTest(unittest.makeSuite(TestCase9, 'test'))
    suite.addTest(unittest.makeSuite(TestCase10, 'test'))
    suite.addTest(unittest.makeSuite(TestParameterSweep1, 'test'))
    suite.addTest(unittest.makeSuite(TestParameterSweep2, 'test'))
    suite.addTest(unittest.makeSuite(TestParameterSweep3, 'test'))
    suite.addTest(unittest.makeSuite(TestParameterSweep4, 'test'))
    suite.addTest(unittest.makeSuite(TestLoop1, 'test'))
    suite.addTest(unittest.makeSuite(TestBranch1, 'test'))
    
    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()




#!/usr/bin/python

import sys

wordcount = {}
for line in sys.stdin.readlines():
    line = line.strip()
    for word in line.split(' '):
        if not len(word):
            continue
        wordcount[word] = wordcount.get(word, 0) + 1

for word, count in wordcount.iteritems():
    print "%s %s" % (word, count)





import os
import sys
import uuid

sys.path.insert(0, os.getenv('POMSETS_HOME'))

import pomsets.context as ContextModule
import pomsets.definition as DefinitionModule
import pomsets.library as LibraryModule
import pomsets.parameter as ParameterModule

import pomsets.test_utils as DefinitionTestModule




def generateBootstrapper():
    defToLoadDef = DefinitionModule.AtomicDefinition()
    defToLoadDef.commandBuilderType('python eval')
    defToLoadDef.id(LibraryModule.ID_LOADLIBRARYDEFINITION)
    defToLoadDef.name('load library definition')
    # need a command builder to call the loadPomset function
    # need a python eval environment to execute the output of commandbuilder
    parameter = ParameterModule.DataParameter(
        id='pomset url', optional=False, active=True,
        portDirection=ParameterModule.PORT_DIRECTION_INPUT)
    ParameterModule.setAttributes(parameter, {})
    defToLoadDef.addParameter(parameter)
    defToLoadDef.isLibraryDefinition(True)
    defToLoadDef.functionToExecute(DefinitionModule.executeTaskInEnvironment)
    return defToLoadDef



def generateDefaultLoader(outputDir):

    definitionsToLoad = []

    defToLoadDef = generateBootstrapper()
    ContextModule.pickleDefinition(
        os.path.join(outputDir, 'loadLibraryDefinition.pomset'), defToLoadDef)
    context = ContextModule.Context()
    context.pomset(defToLoadDef)
    definitionsToLoad.append(context)
    
    wcDefinition = DefinitionTestModule.createWordCountDefinition()
    # wcDefinitionPath = 'wordcount.pomset'
    wcDefinitionPath = wcDefinition.id() + '.pomset'
    wcDefinition.url(wcDefinitionPath)
    ContextModule.pickleDefinition(
        os.path.join(outputDir, wcDefinitionPath), wcDefinition)
    context = ContextModule.Context()
    context.pomset(wcDefinition)
    definitionsToLoad.append(context)
    
    wcrDefinition = DefinitionTestModule.createWordCountReduceDefinition()
    # wcrDefinitionPath = 'wordcount_reduce.pomset'
    wcrDefinitionPath = wcrDefinition.id() + '.pomset'
    wcrDefinition.url(wcrDefinitionPath)
    ContextModule.pickleDefinition(
        os.path.join(outputDir, wcrDefinitionPath), wcrDefinition)
    context = ContextModule.Context()
    context.pomset(wcrDefinition)
    definitionsToLoad.append(context)
    
    library = LibraryModule.Library()
    map(library.addPomsetContext, definitionsToLoad)

    defToLoadDefs = library.generateBootstrapLoaderPomset()
    ContextModule.pickleDefinition(
        os.path.join(outputDir, 'loadLibraryDefinitions.pomset'), defToLoadDefs)
    
    return

def generateLoaderWithFailure1(outputDir):
    
    definitionsToLoad = []

    defToLoadDef = generateBootstrapper()
    ContextModule.pickleDefinition(
        os.path.join(outputDir, 'loadLibraryDefinition.pomset'), defToLoadDef)
    context = ContextModule.Context()
    context.pomset(defToLoadDef)
    definitionsToLoad.append(context)
    
    wcDefinition = DefinitionTestModule.createWordCountDefinition()
    # wcDefinition.id(ID_WORDCOUNT)
    # wcDefinitionPath = 'wordcount.pomset'
    wcDefinitionPath = wcDefinition.id() + '.pomset'
    wcDefinition.url(wcDefinitionPath)
    # we purposely do not pickle it
    # to ensure that the loading fi
    #ContextModule.pickleDefinition(
    #    os.path.join(outputDir, wcDefinitionPath), wcDefinition)
    context = ContextModule.Context()
    context.pomset(wcDefinition)
    definitionsToLoad.append(context)
    
    library = LibraryModule.Library()
    map(library.addPomsetContext, definitionsToLoad)

    defToLoadDefs = library.generateBootstrapLoaderPomset()
    ContextModule.pickleDefinition(
        os.path.join(outputDir, 'loadLibraryDefinitions.pomset'), defToLoadDefs)
    
    return



def generateLoaderWithFailure2(outputDir):

    definitionsToLoad = []

    defToLoadDef = generateBootstrapper()
    ContextModule.pickleDefinition(
        os.path.join(outputDir, 'loadLibraryDefinition.pomset'), defToLoadDef)
    context = ContextModule.Context()
    context.pomset(defToLoadDef)
    definitionsToLoad.append(context)
    
    wcDefinition = DefinitionTestModule.createWordCountDefinition()
    # wcDefinition.id(ID_WORDCOUNT)
    # wcDefinitionPath = 'wordcount.pomset'
    wcDefinitionPath = wcDefinition.id() + '.pomset'
    wcDefinition.url(wcDefinitionPath)
    # we purposely do not pickle it
    # to ensure that the loading fi
    #ContextModule.pickleDefinition(
    #    os.path.join(outputDir, wcDefinitionPath), wcDefinition)
    context = ContextModule.Context()
    context.pomset(wcDefinition)
    definitionsToLoad.append(context)
    
    wcrDefinition = DefinitionTestModule.createWordCountReduceDefinition()
    # wcrDefinition.id(ID_WORDCOUNT_REDUCE)
    # wcrDefinitionPath = 'wordcount_reduce.pomset'
    wcrDefinitionPath = wcrDefinition.id() + '.pomset'
    wcrDefinition.url(wcrDefinitionPath)
    ContextModule.pickleDefinition(
        os.path.join(outputDir, wcrDefinitionPath), wcrDefinition)
    context = ContextModule.Context()
    context.pomset(wcrDefinition)
    definitionsToLoad.append(context)
    
    library = LibraryModule.Library()
    map(library.addPomsetContext, definitionsToLoad)

    defToLoadDefs = library.generateBootstrapLoaderPomset()
    ContextModule.pickleDefinition(
        os.path.join(outputDir, 'loadLibraryDefinitions.pomset'), defToLoadDefs)
    
    return



def main(argv):

    if argv is None:
        argv = []

    shouldGenerateDefaultLoader = True
    shouldGenerateLoaderWithFailure1 = False
    shouldGenerateLoaderWithFailure2 = False
    minArgLen = 2

    for arg in argv:
        if arg.startswith('-default='):
            minArgLen = minArgLen + 1
            shouldGenerateDefaultLoader = arg[9:] in ['True', 'true', '1']
        if arg.startswith('-failure1='):
            minArgLen = minArgLen + 1
            shouldGenerateLoaderWithFailure1 = arg[10:] in ['True', 'true', '1']
        if arg.startswith('-failure2='):
            minArgLen = minArgLen + 1
            shouldGenerateLoaderWithFailure2 = arg[10:] in ['True', 'true', '1']
        pass


    if len(argv) < minArgLen:
        raise ValueError('need to specify directory to output the definitions')
        
    outputDir = argv[-1]

    if shouldGenerateDefaultLoader:
        generateDefaultLoader(outputDir)
        
    if shouldGenerateLoaderWithFailure1:
        generateLoaderWithFailure1(outputDir)

    if shouldGenerateLoaderWithFailure2:
        generateLoaderWithFailure2(outputDir)
    return

if __name__=="__main__":
    main(sys.argv)




from __future__ import with_statement

import logging
import os
import pickle
import uuid

import cloudpool.environment as EnvironmentModule
import cloudpool.task as TaskModule

import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule

import pomsets.context as ContextModule
import pomsets.definition as DefinitionModule
import pomsets.resource as ResourceModule


ID_LOADLIBRARYDEFINITION = 'load library definition::bb028375-bbd5-43ec-b6c3-4955c062063f'
ID_BOOTSTRAPLOADER = 'library bootstrap loader::751fe366-1448-4db3-9db4-944075de7a5b'


def getBootstrapLoaderPomsetsFilter():
    # we need to filter out the bootstrap pomset loaders
    # because it should not be loaded again
    bootstrapLoaderFilter = FilterModule.constructOrFilter()
    bootstrapLoaderFilter.addFilter(
        RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdFilter(ID_LOADLIBRARYDEFINITION)
        )
    )
    bootstrapLoaderFilter.addFilter(
        RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdFilter(ID_BOOTSTRAPLOADER)
        )
    )
    return bootstrapLoaderFilter



class CommandBuilder(TaskModule.CommandBuilder):

    def buildCommand(self, task):
        workRequest = task.workRequest()
        command = 'library.loadFromRelativePath("%s")' % task.getParameterBinding('pomset url')
        
        return command
    
    # END class CommandBuilder
    pass


class LibraryLoader(EnvironmentModule.Environment, ResourceModule.Struct):
    
    DEFAULT_COMMANDBUILDER_TYPE = 'library bootstrap loader'

    ATTRIBUTES = ['library']
    
    def __init__(self, library):
        ResourceModule.Struct.__init__(self)
        self.library(library)
        return
    
    def execute(self, task, *args, **kargs):
        request = task.workRequest()
        
        commandBuilder = self.getCommandBuilder(task)

        command = commandBuilder.buildCommand(task)

        logging.debug("attempt to run %s" % command)
        library = self.library()
        evalResult = eval(command)

        request.kwds['eval result'] = evalResult
        
        return 0
    
    # END class LibraryLoader
    pass


class Path(ResourceModule.Struct):
    
    ATTRIBUTES = ['rawPath']
    
    def __init__(self, rawPath):
        ResourceModule.Struct.__init__(self)
        
        self.rawPath(rawPath)
        pass
    
    # END class Path
    pass


class Library(ResourceModule.Struct):

    BOOTSTRAP_LOADER_FILES = [
        (ID_LOADLIBRARYDEFINITION, 'loadLibraryDefinition.pomset'),
        (ID_BOOTSTRAPLOADER, 'loadLibraryDefinitions.pomset')
    ]
    
    ATTRIBUTES = [
        'hasLoadedDefinitions', 
        'definitionTable',
        'bootstrapLoaderDefinitionsDir',
        'bootstrapLoaderDefinitions',
        'shouldMarkAsLibraryDefinition'
    ]
    
    def __init__(self):
        ResourceModule.Struct.__init__(self)
        self.hasLoadedDefinitions(False)
        self.bootstrapLoaderDefinitions({})
        
        table = RelationalModule.createTable(
            'definitions', 
            ['definition', 'id', 'context'])
        self.definitionTable(table)

        return
    
    def updateWithLibraryDefinitions(self, definition, recursive=True):
        # if the definition contains ReferenceDefinitions
        # and those ReferenceDefinitions indicate that they reference
        # library definitions, 
        if not isinstance(definition, DefinitionModule.CompositeDefinition):
            return
        
        for referenceDefinition in definition.nodes():
            if not referenceDefinition.referencesLibraryDefinition():
                continue
            
            referencedDefinition = referenceDefinition.definitionToReference()
            libraryDefinitionId = referencedDefinition.id()
            
            filter = RelationalModule.ColumnValueFilter(
                'definition',
                FilterModule.IdFilter(libraryDefinitionId))
            
            matchingDefinitions = RelationalModule.Table.reduceRetrieve(
                self.definitionTable(), filter, ['definition'], [])

            if len(matchingDefinitions) is not 0:
                referenceDefinition.definitionToReference(
                    matchingDefinitions[0])
            elif recursive and referencedDefinition is not definition:
                # recursively traverse and update with library definitions
                # but do so only if the reference is not recursive
                self.updateWithLibraryDefinitions(referencedDefinition)
                pass
    
            pass
        
        return


    def loadFromRelativePath(self, relativePath):
        fullPath = os.path.join(
            self.bootstrapLoaderDefinitionsDir(), relativePath)

        return self.loadFromFullFilePath(fullPath)
    

    def loadFromFullFilePath(self, fullPath):
        context = ContextModule.loadPomset(path=fullPath)

        self.addPomsetContext(context)

        return context
    
    
    def loadBootstrapLoaderDefinitions(self):
        """
        This loads the two bootstrapper pomsets
        * the atomic one that loads a single pomset
        * the composite one that specifies a bunch of pomsets to be loaded
        """
        
        dirPath = self.bootstrapLoaderDefinitionsDir()
        if not os.path.exists(dirPath):
            raise NotImplementedError('need to handle when bootstrap loader definitions dir does not exist')
        
        for definitionId, bootstrapLoaderFile in Library.BOOTSTRAP_LOADER_FILES:
            fullPath = os.path.join(dirPath, bootstrapLoaderFile)

            if not os.path.exists(fullPath):
                raise NotImplementedError('need to handle when bootstrap loader file %s does not exist')
            
            pomsetContext = self.loadFromFullFilePath(fullPath)
            pass
        return


    def addPomsetContext(self, context):

        definition = context.pomset()
        definitionId = definition.id()
        if definitionId in [ID_LOADLIBRARYDEFINITION,
                            ID_BOOTSTRAPLOADER]:
            logging.debug("adding definition %s to bootstrap id %s" %
                          (definition, definitionId))

            self.bootstrapLoaderDefinitions()[definitionId] = definition
            pass

        logging.debug("adding definition %s with id %s to library" % 
                      (definition, definitionId))

        if self.shouldMarkAsLibraryDefinition():
            definition.isLibraryDefinition(True)
        self.updateWithLibraryDefinitions(self, definition)

        row = self.definitionTable().addRow()
        row.setColumn('id', definitionId)
        row.setColumn('definition', definition)
        row.setColumn('context', context)
        return


    def hasDefinition(self, filter):
        matchingDefinitions = RelationalModule.Table.reduceRetrieve(
            self.definitionTable(), filter, ['definition'], [])
        return len(matchingDefinitions) is not 0

    
    def removeDefinition(self, filter):
        self.definitionTable().removeRows(filter)
        return

    
    def getDefinition(self, filter):
        matchingDefinitions = RelationalModule.Table.reduceRetrieve(
            self.definitionTable(), filter, ['definition'], [])
        return matchingDefinitions[0]

    
    def getBootstrapLoader(self):
        filter = RelationalModule.ColumnValueFilter(
            'definition',
            FilterModule.IdFilter(ID_BOOTSTRAPLOADER))
        definition = self.getDefinition(filter)
        return definition


    def generateBootstrapLoaderPomset(self):
        """
        generates the composite pomset used to load
        all the library pomsets
        """
        # now create a load library definitions pomset
        # that will load the two wordcount pomsets
        defToLoadDef = self.bootstrapLoaderDefinitions()[ID_LOADLIBRARYDEFINITION]

        defToLoadDefs = DefinitionModule.getNewNestDefinition()

        notBootstrapLoaderFilter = FilterModule.constructNotFilter()
        notBootstrapLoaderFilter.addFilter(getBootstrapLoaderPomsetsFilter())
        
        definitions = RelationalModule.Table.reduceRetrieve(
            self.definitionTable(), 
            notBootstrapLoaderFilter,
            ['definition'], [])
        
        for definitionToLoad in definitions:
            loadNode = defToLoadDefs.createNode(id=uuid.uuid4())
            loadNode.definitionToReference(defToLoadDef)
            loadNode.isCritical(False)
            loadNode.name('load %s' % definitionToLoad.url())
            loadNode.setParameterBinding('pomset url', definitionToLoad.url())
            pass
        
        defToLoadDefs.id(ID_BOOTSTRAPLOADER)
        defToLoadDefs.name('bootstrap pomsets loader')

        return defToLoadDefs

    
    def saveBootstrapLoaderPomset(self, outputPath=None):
        
        # default to the library's specified dir
        if outputPath is None:
            outputPath = os.path.join(
                self.bootstrapLoaderDefinitionsDir(),
                'loadLibraryDefinitions.pomset')

        pomset = self.generateBootstrapLoaderPomset()
        
        ContextModule.pickleDefinition(outputPath, pomset)

        return

    
    # END class Library
    pass




import logging
import os


import pomsets.builder as BuilderModule
import pomsets.command as CommandModule
import pomsets.definition as DefinitionModule


def getHomeLocation():
    return os.getenv('HADOOP_HOME') 

def getExecutablePath():
    return os.path.join(getHomeLocation(), 'bin', 'hadoop')

def getStreamingJar():
    return os.path.join(getHomeLocation(), 'streaming.jar')

class JarExecutable(CommandModule.Executable):

    # TODO:
    # in the future, whether something is stageable
    # will be determined by whether its dependencies
    # are all stageable
    ATTRIBUTES = CommandModule.Executable.ATTRIBUTES + [
        'jarFile',
        'jarClass'
    ]
    
    def __init__(self):
        CommandModule.Executable.__init__(self)
        self.jarClass([])
        self.jarFile([])
        return
    
    def staticArgs(self, value=None):
        return ['jar'] + self.jarFile() + self.jarClass()
    
    # END class Executable
    pass


class PipesExecutable(CommandModule.Executable):
    
    # TODO:
    # in the future, whether something is stageable
    # will be determined by whether its dependencies
    # are all stageable
    ATTRIBUTES = CommandModule.Executable.ATTRIBUTES + [
        'pipesFile',
    ]
    
    def __init__(self):
        CommandModule.Executable.__init__(self)
        self.pipesFile([])
        return
    
    def staticArgs(self, value=None):
        return ['pipes', '-program'] + self.pipesFile()
    
    # END class PipesExecutable
    pass



class Builder(BuilderModule.Builder):


    def createExecutableObject(self, jarFile, jarClass):
        executable = JarExecutable()
        executable.stageable(False)
        executable.path([getExecutablePath()])
        executable.jarFile([jarFile])
        executable.jarClass([jarClass])

        return executable


    def createStreamingExecutableObject(self):

        executable = JarExecutable()
        executable.stageable(False)
        executable.path([getExecutablePath()])
        executable.jarFile([getStreamingJar()])
        executable.jarClass([])

        return executable


    def createPipesExecutableObject(self, path, staticArgs=None):
        command = [path]
        executable = PipesExecutable()
        executable.stageable(False)
        executable.path([getExecutablePath()])
        executable.pipesFile(command)
        return executable


    # END class Builder
    pass



from __future__ import with_statement

import os
import sys
import threadpool
import unittest
import logging
import uuid

import currypy

import pomsets.automaton as AutomatonModule
import pypatterns.filter as FilterModule

import pomsets.command as TaskCommandModule
import pomsets.definition as DefinitionModule
import pomsets.parameter as ParameterModule
import pomsets.task as TaskModule


class TestParameterOrder(unittest.TestCase):
    
    def testOrder1(self):
        """
        two parameters with no specified ordering returns 0
        """
        raise NotImplementedError
    
    def testOrder2(self):
        """
        two parameters with ordering specified returns either 1 or -1
        """
        raise NotImplementedError
    
    def testOrder3(self):
        """
        two parameters that both succeed another parameter 
        but have no relations to each other returns 0
        """
        raise NotImplementedError

    def testOrder4(self):
        """
        two parameters that both precede another parameter 
        but have no relations to each other returns 0
        """
        raise NotImplementedError

    def testOrder5(self):
        """
        two parameters that relate to each other transitively
        but have no direct relations to each other returns either 1 or -1
        """
        raise NotImplementedError
    
    # END class TestParameterOrder
    pass


class TestParameterArgument(unittest.TestCase):

    def assertParameterArgument(self, parameter, parameterBindings, expected):
        argument = TaskCommandModule.buildParameterArgument(
            parameter, parameterBindings)
        assert expected == argument, \
               'expected argument to be %s, got %s' % (expected, argument)
        return
    
        
    def testBasic(self):
        """
        tests that the parameter's value gets output
        """

        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = ['foo']
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     parameterValue)
        return
    
    def testBasicFlag(self):
        """
        tests that a flag gets output 
        """
        
        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        prefixFlag = ['-input']
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:prefixFlag
                    },        
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = ['foo']
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     prefixFlag + parameterValue)
        return
    
    
    
    def testBooleanFlagImplicitTrue(self):
        """
        tests for boolean value, implicitly specified by the flag,
        e.g. "-foo" is True and "" is False
        """
        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)

        trueValue = ['-foo']
        falseValue = []
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_ISENUM:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_ENUM_MAP:{
                        True:trueValue,
                        False:falseValue
                        }
                    },
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = [True]
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     trueValue)

        return

    
    def testBooleanFlagImplicitTrue(self):
        """
        tests for boolean value, implicitly specified by the flag,
        e.g. "-foo" is True and "" is False
        """
        # create the parameter
        parameterId = uuid.uuid4().hex
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterId, name=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)

        trueValue = ['-foo']
        falseValue = []
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_ISENUM:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_ENUM_MAP:{
                        True:trueValue,
                        False:falseValue
                        }
                    },
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = [False]
        parameterBindings = {
            parameterId:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     falseValue)

        return
    
    
    
    def testBooleanFlagExplicitTrue(self):
        """
        tests for boolean value, explicity specified by the value
        and not by the flag, e.g.
        "-foo True" and "-foo False"
        """
        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)

        prefixFlag = ['-foo']
        trueValue = ['1']
        falseValue = ['0']
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_ISENUM:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:prefixFlag,
                    ParameterModule.COMMANDLINE_ENUM_MAP:{
                        True:trueValue,
                        False:falseValue
                        }
                    },
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = [True]
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     prefixFlag + trueValue)
        
        return

    
    def testBooleanFlagExplicitFalse(self):
        """
        tests for boolean value, explicity specified by the value
        and not by the flag, e.g.
        "-foo True" and "-foo False"
        """
        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)

        prefixFlag = ['-foo']
        trueValue = ['1']
        falseValue = ['0']
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_ISENUM:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:prefixFlag,
                    ParameterModule.COMMANDLINE_ENUM_MAP:{
                        True:trueValue,
                        False:falseValue
                        }
                    },
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = [False]
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     prefixFlag + falseValue)
        
        return
    
    
    def testBasicListValue(self):
        """
        tests to ensure that a list of values is output
        """
        
        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = ['foo', 'bar']
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     parameterValue)
        return

    
    def testListValueWithFlag(self):
        """
        tests to ensure that a list of values is output
        along with its prefix flag (non-distributed)
        """
        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        prefixFlag = ['-input']
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:prefixFlag
                    },        
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = ['foo', 'bar']
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(parameter,
                                     parameterBindings,
                                     prefixFlag + parameterValue)
        return


    def testListValueWithFlagDistributed(self):
        """
        tests to ensure that a list of values is output
        along with its prefix flag (distributed)
        """
        # create the parameter
        parameterName = 'test'
        parameter = ParameterModule.DataParameter(
            id=parameterName, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        prefixFlag = ['-input']
        parameterAttributes = {
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:True,
                ParameterModule.PORT_ATTRIBUTE_ISINPUTFILE:True,
                ParameterModule.PORT_ATTRIBUTE_ISLIST:True,
                ParameterModule.PORT_ATTRIBUTE_COMMANDLINE_OPTIONS:{
                    ParameterModule.COMMANDLINE_PREFIX_FLAG:prefixFlag,
                    ParameterModule.COMMANDLINE_PREFIX_FLAG_DISTRIBUTE:True
                    },        
                }
        ParameterModule.setAttributes(parameter, parameterAttributes)

        # create the parameter bindings
        parameterValue = ['foo', 'bar']
        parameterBindings = {
            parameterName:parameterValue
        }
                
        # verify the parameter argument
        self.assertParameterArgument(
            parameter,
            parameterBindings,
            reduce(lambda x, y: x+prefixFlag+[y],
                   parameterValue, [])
        )

        return
    
    
    # END class TestParameterArgument
    pass




def main():
    utils.configLogging()

    suite = unittest.TestSuite()
    
    suite.addTest(unittest.makeSuite(TestParameterArgument, 'test'))
    
    runner = unittest.TextTestRunner()
    runner.run(suite)
    return

if __name__=="__main__":
    main()




import copy
import logging
import subprocess
import time

import pypatterns.filter as FilterModule
import pomsets.graph as GraphModule
import pypatterns.relational as RelationalModule
import pomsets.resource as ResourceModule

import pomsets.parameter as ParameterModule


def doTask(task, *args, **kwds):
    kwds['task'] = task
    kwds['thread pool'] = task.threadPool()
    return task.do()


class GuiOptionsHolder(ResourceModule.Struct):

    ATTRIBUTES = ['guiOptions']

    # END class GuiOptionsHolder
    pass

class ParameterBindingsHolder(ResourceModule.Struct):

    ATTRIBUTES = ['parameterBindings']

    def __init__(self):
        ResourceModule.Struct.__init__(self)
        self.parameterBindings({})
        pass

    def setParameterBinding(self, key, value):
        logging.debug(
            'setting key "%s" to values "%s" to %s\'s parameter bindings' % 
            (key, value, self))       

        self.parameterBindings()[key] = value
    
    def getParameterBinding(self, key):
        if not self.hasParameterBinding(key):
            raise KeyError('%s not in %s\'s parameter bindings' % (key, self))
        return self.parameterBindings()[key]
        
    def hasParameterBinding(self, key):
        return key in self.parameterBindings()

    # END class ParameterBindingsHolder
    pass


class Definition(ResourceModule.Struct):
    
    ATTRIBUTES = GuiOptionsHolder.ATTRIBUTES + [
        'id',
        'parametersTable',
        'isLibraryDefinition',
        'url',
        'description',
        'name',
        'parameterOrderingTable',
    ]

    SYMBOL_INPUT = 'input'
    SYMBOL_OUTPUT = 'output'
    SYMBOL_INPUT_TEMPORAL = 'temporal input'
    SYMBOL_OUTPUT_TEMPORAL = 'temporal output'
    
    def __init__(self, id=None, name=None):
        ResourceModule.Struct.__init__(self)

        if id is None:
            id = ResourceModule.Resource.ID_GENERATOR()
        self.id(id)
        
        if name is None:
            name = ''
        self.name(name)
        
        self.initializeParameters()
        self.isLibraryDefinition(False)
        return
    
    
    def addParameter(self, parameter, id=None):

        row = self.parametersTable().addRow()

        # TODO:
        # why do we allow of a different id
        # than the one of the parameter?
        if id is None:
            id = parameter.id()
        row.setColumn('id', id)

        row.setColumn('parameter', parameter)
        row.setColumn('port direction', parameter.portDirection())
        row.setColumn('port type', parameter.portType())
        
        return parameter
 
    
    def removeParameter(self, parameter):
        filter = RelationalModule.ColumnValueFilter(
            'parameter',
            FilterModule.IdentityFilter(parameter)
        )
        self.parametersTable().removeRows(filter)
        return

    
    def initializeParameters(self):
        
        table = RelationalModule.createTable(
            'parameters', 
            ['id', 'parameter', 'port direction', 'port type']
        )
        self.parametersTable(table)
        
        inputParameter = ParameterModule.InputTemporalParameter(
            Definition.SYMBOL_INPUT_TEMPORAL)
        
        outputParameter = ParameterModule.OutputTemporalParameter(
            Definition.SYMBOL_OUTPUT_TEMPORAL)
        
        self.addParameter(inputParameter)
        self.addParameter(outputParameter)
        return

    def getParametersByFilter(self, filter):
        theFilter = RelationalModule.ColumnValueFilter(
            'parameter',
            filter)
        
        parameters = RelationalModule.Table.reduceRetrieve(
            self.parametersTable(),
            theFilter,
            ['parameter'],
            []
        )
        
        return parameters

    def getParametersHavingId(self, id):
        filter = RelationalModule.ColumnValueFilter(
            'id',
            FilterModule.EquivalenceFilter(id)
        )
        
        parameters = RelationalModule.Table.reduceRetrieve(
            self.parametersTable(),
            filter,
            ['parameter']
        )
        return parameters
    
    
    def hasParameter(self, id):
        parameters = self.getParametersHavingId(id)
        return len(parameters) is not 0
    
    def getParameter(self, id):
        parameters = self.getParametersHavingId(id)
        
        if len(parameters) is 0:
            raise NotImplementedError(
                '%s has no parameter with id %s' % (self.name(), id))
        elif len(parameters) is not 1:
            raise NotImplementedError(
                '%s has more than one parameter with id %s' % self.name(), id)
        return parameters[0]

    
    def isAtomic(self):
        return False
    
    pass


class CompositeDefinition(GraphModule.Graph, Definition, 
                          ParameterBindingsHolder):
    
    ATTRIBUTES = GraphModule.Graph.ATTRIBUTES + \
               Definition.ATTRIBUTES + \
               ParameterBindingsHolder.ATTRIBUTES + \
               ['definitions', 'parameterConnectionsTable',
                'parameterConnectionPathTable']

    
    def __init__(self):
        GraphModule.Graph.__init__(self)
        Definition.__init__(self)

        ParameterBindingsHolder.__init__(self)

        # definitions are 
        # the "alphabet" of the pomset
        self.definitions(set([]))
        
        self.initializeParameterConnections()
        self.initializeParameterConnectionPaths()
        return
    
    def __eq__(self, other):
        differences = [x for x in self.getDifferencesWith(other)]
        return len(differences) is not 0
    
    def getDifferencesWith(self, other):
        if not other:
            yield 'other is None'
            raise StopIteration
        if not self.__class__ == other.__class__:
            yield 'other is of class %s' % other.__class__
            raise StopIteration

        # parameters
        selfParameters = dict(
            [(x.id(), x) 
             for x in self.getParametersByFilter(FilterModule.TRUE_FILTER)]
        )
        otherParameters = dict(
            [(x.id(), x) 
             for x in other.getParametersByFilter(FilterModule.TRUE_FILTER)]
        )
        diffParameters = \
            set(selfParameters.keys()).symmetric_difference(otherParameters.keys())
        if len(diffParameters) is not 0:
            yield 'different number of parameters'
        else:
            for parameterId, selfParameter in selfParameters.iteritems():
                otherParameter = otherParameters[parameterId]
                if not selfParameter == otherParameter:
                    yield 'parameter %s not the same' % parameterId
                    
        
        # nodes
        selfNodes = dict(
            [(x.id(), x) for x in self.nodes()]
        )
        otherNodes = dict(
            [(x.id(), x) for x in other.nodes()]
        )
        diffNodes = \
            set(selfNodes.keys()).symmetric_difference(otherNodes.keys())
        if len(diffNodes) is not 0:
            yield 'different number of nodes'
        else:
            for nodeId, selfNode in selfNodes.iteritems():
                otherNode = otherNodes[nodeId]
                if not selfNode == otherNode:
                    yield 'node %s not the same' % nodeId
                    

        # parameter connections
        columns = ['source node', 'source parameter', 
                   'target node', 'target parameter']
        for row in self.parameterConnectionsTable().retrieve(columns=columns):
            # TODO:
            # should compare parameter connections
            
            pass
            
        # parameter bindings
        if not self.parameterBindings() == other.parameterBindings():
            yield 'parameter bindings'
            
        raise StopIteration
    
    def __copy__(self):
        
        result = self.__class__()
        
        memo = {self:result}

        # copy the name
        result.name(self.name())

        # copy over the parameters
        notTemporalFilter = FilterModule.constructNotFilter()
        notTemporalFilter.addFilter(
            FilterModule.EquivalenceFilter(ParameterModule.PORT_TYPE_TEMPORAL)
        )
        parameterFilter = FilterModule.ObjectKeyMatchesFilter(
            keyFunction = lambda x: x.portType(),
            filter = notTemporalFilter
        )
        map(result.addParameter, self.getParametersByFilter(parameterFilter))
            
        # copy over the reference definitions
        for node in self.nodes():
            newNode = copy.copy(node)
            result.addNode(newNode)
            memo[node] = newNode
            
        # copy over the parameter connections
        columns = ['source node', 'source parameter', 
                   'target node', 'target parameter']
        for row in self.parameterConnectionsTable().retrieve(columns=columns):
            sourceNode = row[0]
            targetNode = row[2]
            sourceParameter = row[1]
            targetParameter = row[3]
            copySourceNode = memo[sourceNode]
            copyTargetNode = memo[targetNode]
            connection = result.connectParameters(
                copySourceNode, sourceParameter,
                copyTargetNode, targetParameter
                )
            result.addParameterConnectionPath(
                copySourceNode, sourceParameter,
                copyTargetNode, targetParameter,
                tuple([connection])
                )

            pass
            
        # copy over the parameter bindings
        result.parameterBindings(self.parameterBindings())
        
        return result
    
    def getClassForCreateNode(self):
        return ReferenceDefinition
    
    def getClassForCreateEdge(self):
        return ParameterModule.ParameterConnection


    def initializeParameterConnectionPaths(self):
        table = RelationalModule.createTable(
            'parameter connection path',
            ['source node', 'source parameter', 
             'target node', 'target parameter',
             'path', 'additional parameters']
        )
        self.parameterConnectionPathTable(table)
        return

    def initializeParameterConnections(self):
        table = RelationalModule.createTable(
            'parameter connections',
            ['source node', 'source parameter', 
             'target node', 'target parameter',
             'parameter connection']
        )
        self.parameterConnectionsTable(table)
        return
    

    def addDefinition(self, definition):
        self.definitions().add(definition)
        return

    def removeDefinition(self, definition):
        # first need to verify that there is no node
        # which references the definition
        # if there is, throw an exception
        referencedDefinitions = [x.definitionToReference() for x in self.nodes()]
        if definition in referencedDefinitions:
            raise ValueError('cannot remove referenced definition')
        if definition in self.definitions():
            self.definitions().remove(definition)
        return

    def removeNode(self, node):

        # we need to remove the edges here
        # instead of defaulting to grpah
        # because what's presented as an data edge to the user
        # is actually two parameter connections
        # though temporal connections are just a single connection

        # remove just the node, 
        # since we've removed the edges
        GraphModule.Graph.removeNode(self, node, shouldRemoveEdges=False)

        return
    
    def addNewNode(self, definitionToReference, 
                   predecessors=None, successors=None):

        if predecessors is None:
            predecessors = []
        if successors is None:
            successors = []

        # add definitionToReference of definitions that
        # this definition references
        self.addDefinition(definitionToReference)
        
        id = ResourceModule.Resource.ID_GENERATOR()
        node = self.createNode(id=id)

        node.definitionToReference(definitionToReference)

        # add temporal connections
        for predecessor in predecessors:
            connection = self.connectParameters(
                predecessor, 'temporal output',
                node, 'temporal input'
            )
            self.addParameterConnectionPath(
                predecessor, 'temporal output',
                node, 'temporal input',
                tuple([connection])
                )
            pass
        for successor in successors:
            connection = self.connectParameters(
                node, 'temporal output',
                successor, 'temporal input'
            )
            self.addParameterConnectionPath(
                node, 'temporal output',
                successor, 'temporal input',
                tuple([connection])
                )
            pass

        return node
    
    
    def connectParameters(self, 
                          sourceNode, sourceParameter, 
                          targetNode, targetParameter):
        
        edge = self.createEdge([sourceNode, targetNode])
        # this sets the references for parameter connections
        edge.setReferences(sourceNode, sourceParameter,
                           targetNode, targetParameter)

        row = self.parameterConnectionsTable().addRow()
        row.setColumn('source node', sourceNode)
        row.setColumn('source parameter', sourceParameter)
        row.setColumn('target node', targetNode)
        row.setColumn('target parameter', targetParameter)
        row.setColumn('parameter connection', edge)
        
        return edge


    def addParameterConnectionPath(self, 
                                   sourceNode, sourceParameter,
                                   targetNode, targetParameter, path,
                                   additionalParameters=None):

        if additionalParameters is None:
            additionalParameters = tuple([])

        row = self.parameterConnectionPathTable().addRow()

        row.setColumn('source node', sourceNode)
        row.setColumn('source parameter', sourceParameter)
        row.setColumn('target node', targetNode)
        row.setColumn('target parameter', targetParameter)
        row.setColumn('path', path)
        row.setColumn('additional parameters', additionalParameters)

        return



    def constructParameterConnectionFilter(self, 
                                           sourceNode, sourceParameterId,
                                           targetNode, targetParameterId):
        filter = FilterModule.constructAndFilter()
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'source node',
                FilterModule.IdentityFilter(sourceNode)
                )
            )
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'source parameter',
                FilterModule.EquivalenceFilter(sourceParameterId)
                )
            )
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target node',
                FilterModule.IdentityFilter(targetNode)
                )
            )
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target parameter',
                FilterModule.EquivalenceFilter(targetParameterId)
                )
            )
        return filter


    def removeParameterConnection(self, connection):
        filter = self.constructParameterConnectionFilter(
            connection.sourceNode(),
            connection.sourceParameter(),
            connection.targetNode(),
            connection.targetParameter())
        self.parameterConnectionsTable().removeRows(filter)
        self.removeEdge(connection)
        return



    def getIdForParameterReference(self, node, parameterId):
        if node is self:
            return parameterId
        return '%s.%s' % (node.id(), parameterId)
    
    
    def _getNotSelfParameterConnectionFilter(self):
    
        theSelfFilter = FilterModule.constructOrFilter()
        theSelfFilter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter = FilterModule.IdentityFilter(self),
                keyFunction = lambda x: x.sourceNode()
            )
        )
        theSelfFilter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter = FilterModule.IdentityFilter(self),
                keyFunction = lambda x: x.targetNode()
            )
        )
        theNotSelfFilter = FilterModule.constructNotFilter()
        theNotSelfFilter.addFilter(theSelfFilter)
        return theNotSelfFilter
        
    
    def getMinimalNodes(self):
        theNotSelfFilter = self._getNotSelfParameterConnectionFilter()
        edges = [x for x in theNotSelfFilter.wrap(self.edges())]
        targetNodes = set([x.targetNode() for x in edges])
        allNodes = set(self.nodes())
        
        return allNodes.difference(targetNodes)

    def getMaximalNodes(self):
        theNotSelfFilter = self._getNotSelfParameterConnectionFilter()
        edges = [x for x in theNotSelfFilter.wrap(self.edges())]
        sourceNodes = set([x.sourceNode() for x in edges])
        allNodes = set(self.nodes())
        
        return allNodes.difference(sourceNodes)


    
    def functionToExecute(self):
        return doTask
    
    # END class CompositeDefinition
    pass
    

class NestDefinition(CompositeDefinition):
    
    ATTRIBUTES = CompositeDefinition.ATTRIBUTES


    pass


class LoopDefinition(CompositeDefinition):

    ATTRIBUTES = CompositeDefinition.ATTRIBUTES
    
    PARAMETER_INITIAL_STATE = 'loop initial state'
    PARAMETER_STATE = 'loop state'
    PARAMETER_CONTINUE_CONDITION = 'loop continue condition'
    PARAMETER_STATE_TRANSITION = 'loop state transition'
    PARAMETER_STATE_CONFIGURATION = 'loop state configuration'

    def __init__(self):
        CompositeDefinition.__init__(self)
        
        parameter = ParameterModule.DataParameter(
            id=LoopDefinition.PARAMETER_INITIAL_STATE, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        self.addParameter(parameter)
        
        parameter = ParameterModule.DataParameter(
            id=LoopDefinition.PARAMETER_CONTINUE_CONDITION, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        self.addParameter(parameter)
        
        parameter = ParameterModule.DataParameter(
            id=LoopDefinition.PARAMETER_STATE_TRANSITION, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        self.addParameter(parameter)

        parameter = ParameterModule.DataParameter(
            id=LoopDefinition.PARAMETER_STATE_CONFIGURATION, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        self.addParameter(parameter)
        
        parameter = ParameterModule.BlackboardParameter(
            LoopDefinition.PARAMETER_STATE)
        self.addParameter(parameter)
        
        connection = self.connectParameters(
            self, LoopDefinition.PARAMETER_INITIAL_STATE,
            self, LoopDefinition.PARAMETER_STATE
        )
        self.addParameterConnectionPath(
            self, LoopDefinition.PARAMETER_INITIAL_STATE,
            self, LoopDefinition.PARAMETER_STATE,
            tuple([connection])
            )
        
        return
    
    
    # END class LoopDefinition
    pass

class BranchDefinition(CompositeDefinition):
    
    ATTRIBUTES = CompositeDefinition.ATTRIBUTES
    
    PARAMETER_CONDITION_STATE = 'branch condition state'
    PARAMETER_CONDITION_FUNCTION = 'branch condition function'
    PARAMETER_CONDITION_MAP = 'branch condition map'
    
    def __init__(self):
        CompositeDefinition.__init__(self)
        
        parameter = ParameterModule.DataParameter(
            id=BranchDefinition.PARAMETER_CONDITION_STATE, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        self.addParameter(parameter)

        parameter = ParameterModule.DataParameter(
            id=BranchDefinition.PARAMETER_CONDITION_FUNCTION, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        self.addParameter(parameter)

        parameter = ParameterModule.DataParameter(
            id=BranchDefinition.PARAMETER_CONDITION_MAP, 
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        self.addParameter(parameter)

        
        return
    
    # END class BranchDefinition
    pass

class AtomicDefinition(Definition):

    ATTRIBUTES = Definition.ATTRIBUTES + [
        'executable',
        'functionToExecute',
        'commandBuilderType'
    ]
    
    def __init__(self, **kwds):
        Definition.__init__(self, **kwds)
        pass
    
    def hasParameterBinding(self, key):
        return False
    
    def isAtomic(self):
        return True
    
    # END class AtomicDefinition
    pass



def executeTaskInEnvironment(task, *args, **kwds):
    environment = task.workRequest().kwds['execute environment']
    return environment.execute(task, *args, **kwds)


def createParameterOrderingTable():
    table = RelationalModule.createTable('parameter ordering', 
                                         ['source', 'target'])
    return table
    

def createPythonEvalDefinition():
    definition = AtomicDefinition()
    definition.commandBuilderType('python eval')
    
    parameter = ParameterModule.DataParameter(
        id='command', optional=False, active=True,
        portDirection=ParameterModule.PORT_DIRECTION_INPUT)
    ParameterModule.setAttributes(parameter, {})
    definition.addParameter(parameter)
    
    parameter = ParameterModule.DataParameter(
        id='eval result', optional=False, active=True,
        portDirection=ParameterModule.PORT_DIRECTION_OUTPUT)
    ParameterModule.setAttributes(parameter, {})
    definition.addParameter(parameter)
    
    return definition


def createShellProcessDefinition(inputParameters=None,
                                 parameterOrderings=None,
                                 executable=None,
                                 definitionClass=None,
                                 *args, **kwds):
    """
    TODO:
    move this into a utility file
    that makes use of both the definition and command modules
    
    @param inputParameters- a dict of input parameter id to boolean 
           indicating whether they are commandline args
    @param parameterOrderings- a table that specifies the ordering of parameters
           will be used by the comparator for parameter sorting
    """
    
    # this is to handle the case
    # where we are just executing the "command" input
    # as a single command, and not trying to combine commandlines together
    if inputParameters is None:
        inputParameters = {
            'command':{ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:False}
        }
        
    if definitionClass is None:
        definitionClass = AtomicDefinition
    definition = definitionClass(*args, **kwds)
    definition.functionToExecute(executeTaskInEnvironment)

    for inputParameterName, parameterAttributes in inputParameters.iteritems():
        
        isOptional = False
        if parameterAttributes.get(ParameterModule.PORT_ATTRIBUTE_COMMANDLINE, False):
            isOptional = parameterAttributes.get(ParameterModule.PORT_ATTRIBUTE_ISOPTIONAL, False)
            pass

        # whether something is active is actually dynamically set
        # and should be specified on the reference definition
        isActive = True

        parameter = ParameterModule.DataParameter(
            id=inputParameterName, 
            optional=isOptional, active=isActive,
            portDirection=ParameterModule.PORT_DIRECTION_INPUT)
        
        ParameterModule.setAttributes(parameter, parameterAttributes)
        
        definition.addParameter(parameter)
        pass

    if parameterOrderings is None:
        parameterOrderings = createParameterOrderingTable()
    # TODO:
    # need to compute the full transitive reflexive closure of precedences
    definition.parameterOrderingTable(parameterOrderings)

    # default the executable if necessary
    if executable is None:
        import pomsets.command as ExecuteCommandModule
        executable = ExecuteCommandModule.Executable()
        executable.stageable(False)
        executable.path([])
        executable.staticArgs([])
        
    # set the executable
    definition.executable(executable)
    
    definition.commandBuilderType('shell process')
    
    return definition


DEFINITION_SHELLPROCESS = createShellProcessDefinition(
    inputParameters={'command':{ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:False}}
    )
DEFINITION_SSHPROCESS = createShellProcessDefinition(
    inputParameters={'command':{ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:False},
                     'connection':{ParameterModule.PORT_ATTRIBUTE_COMMANDLINE:False}}
    )


    

class ReferenceDefinition(GraphModule.Node, ParameterBindingsHolder):
    
    ATTRIBUTES = GraphModule.Node.ATTRIBUTES + \
               ParameterBindingsHolder.ATTRIBUTES + \
               GuiOptionsHolder.ATTRIBUTES + [
        'parameterSweeps',
        'parameterSweepGroups',
        'parameterStagingMap',
        'comment',
        'isCritical'
    ]
    
    def __init__(self, id=None, graph=None):
        GraphModule.Node.__init__(self, id=id, graph=graph)

        ParameterBindingsHolder.__init__(self)

        self.parameterSweeps(set([]))
        
        table = RelationalModule.createTable('parameter sweep groups',
                                             ['id', 'group'])
        self.parameterSweepGroups(table)

        # this allows us to persist GUI options
        # e.g. node position
        self.guiOptions({})
        
        self.parameterStagingMap({})

        # default to True
        self.isCritical(True)

        pass

    """
    def __eq__(self, other):
        differences = [x for x in self.getDifferencesWith(other)]
        return len(differences) is 0
    
    def getDifferencesWith(self, other):
        if not other:
            yield 'other is None'
            raise StopIteration
        if not self.__class__ == other.__class__:
            yield 'other is of class %s' % other.__class__
            raise StopIteration

        if not self.definitionToReference() == other.definitionToReference():
            yield 'referenced definition'
        
        if not self.parameterBindings() == other.parameterBindings():
            yield 'parameter bindings'
        
        if not self.parameterSweeps() == other.parameterSweeps():
            yield 'parameter sweeps'

        if not self.parameterSweepGroups() == other.parameterSweepGroups():
            yield 'parameter sweep groups'

        if not self.parameterStagingMap() == other.parameterStagingMap():
            yield 'parameter staging map'
        
        return
    """
    def __eq__(self, other):
        return self is other


    def __copy__(self):
        result = ReferenceDefinition(id=self.id())
        result.name(self.name())
        
        result.definitionToReference(self.resource())
        
        result.parameterBindings(self.parameterBindings())
        result.parameterSweeps(self.parameterSweeps())
        result.parameterSweepGroups(self.parameterSweepGroups())
        result.guiOptions(copy.copy(self.guiOptions()))
        result.parameterStagingMap(self.parameterStagingMap())
        
        return result
    
    
    def isAtomic(self):
        if isinstance(self.definitionToReference(), CompositeDefinition):
            return False
        if self.hasParameterSweep():
            return False
        return True
            
    def executable(self):
        return self.definitionToReference().executable()
    
    
    def functionToExecute(self, value=None):
        if value is not None:
            # we do not allow the function to be set
            raise NotImplementedError
        return self._functionToExecute

    def parameterOrderingTable(self):
        return self.definitionToReference().parameterOrderingTable()
    
    def getParameter(self, id):
        return self.definitionToReference().getParameter(id)
    
    
    def getParametersByFilter(self, filter):
        """
        # we've commented this out
        # because we don't want to just return the unbound parameters
        # don't remove this, as this code should go elsewhere
        
        theFilter = FilterModule.constructAndFilter()
        
        theBoundParameterFilter = FilterModule.constructOrFilter()
        for key in self.parameterBindings().keys():
            theBoundParameterFilter.addFilter(
                FilterModule.IdFilter(key)
            )
            pass
        theNotBoundParameterFilter = FilterModule.constructNotFilter()
        theNotBoundParameterFilter.addFilter(theBoundParameterFilter)
        
        theFilter.addFilter(filter)
        theFilter.addFilter(theNotBoundParameterFilter)
        
        return self.definitionToReference().getParametersByFilter(theFilter)
        """
        return self.definitionToReference().getParametersByFilter(filter)
    
    
    def getParameterToEdit(self, parameterName):

        parameter = self.getParameter(parameterName)

        logging.debug('parameter is of type %s' % parameter.portType())

        if parameter.portType() == ParameterModule.PORT_TYPE_BLACKBOARD:
            return (self, parameter)
        if parameter.portType() == ParameterModule.PORT_TYPE_TEMPORAL:
            return (self, parameter)

        logging.debug('parameter is of direction %s' % 
                      parameter.portDirection())
        if parameter.portDirection() == ParameterModule.PORT_DIRECTION_OUTPUT:
            return (self, parameter)
        if parameter.portDirection() == ParameterModule.PORT_DIRECTION_INTERNAL:
            return (self, parameter)

        # at this point, we should only be left with input data parameters
        # so, if there is an incoming parameter connection
        # and the source of that parameter connection is a blackboard parameter
        # then that is the parameter to edit
        """
        parameterConnectionFilter = FilterModule.constructAndFilter()
        parameterConnectionFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'source node',
                FilterModule.IdentityFilter(self.graph())
                )
            )
        parameterConnectionFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target node',
                FilterModule.IdentityFilter(self)
                )
            )
        parameterConnectionFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target parameter',
                FilterModule.EquivalenceFilter(parameterName)
                )
            )

        rows = [x for x in 
                self.graph().parameterConnectionsTable().retrieve(
                parameterConnectionFilter, 
                ['source node', 'source parameter']
                )]
        parameters = []
        for sourceNode, sourceParameterName in rows:
            sourceParameter = sourceNode.getParameter(sourceParameterName)
            if not sourceParameter.portType() ==  ParameterModule.PORT_TYPE_BLACKBOARD:
                continue
            parameters.append((sourceNode, sourceParameter))
        """
        filter = FilterModule.constructAndFilter()
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target node',
                FilterModule.IdentityFilter(self)
                )
            )
        filter.addFilter(
            RelationalModule.ColumnValueFilter(
                'target parameter',
                FilterModule.EquivalenceFilter(parameterName)
                )
            )
        
        parameters = RelationalModule.Table.reduceRetrieve(
            self.graph().parameterConnectionPathTable(),
            filter,
            ['additional parameters'], [])


        if len(parameters) is 0:
            logging.debug('no incoming blackboard parameters found')
            return (self, parameter)

        if len(parameters) is not 1:
            raise ValueError('expected only one incoming connection from a blackboard parameter, got %s' % len(parameters))

        parameterId = parameters[0][0]
        parameter = self.graph().getParameter(parameterId)

        logging.debug('returning %s as the parameter to edit for (%s,%s)' %
                      (parameter, self, parameterName))

        # parameters is a list of tuples of (node, parameter)
        return (self.graph(), parameter)

    
    def referencesLibraryDefinition(self):
        definition = self.definitionToReference()
        if definition is None:
            return False
        return definition.isLibraryDefinition()
    
    
    def definitionToReference(self, definition=None):
        if definition is not None:
            self._functionToExecute = definition.functionToExecute()
            pass
        return self.resource(definition)

    
    def incomingEdgeFilter(self):
        filter = FilterModule.constructAndFilter()
        filter.addFilter(self.graph()._getNotSelfParameterConnectionFilter())
        filter.addFilter(GraphModule.Node.incomingEdgeFilter(self))
        return filter

    def outgoingEdgeFilter(self):
        filter = FilterModule.constructAndFilter()
        filter.addFilter(self.graph()._getNotSelfParameterConnectionFilter())
        filter.addFilter(GraphModule.Node.outgoingEdgeFilter(self))
        return filter
    
    def parameterConnectionsTable(self):
        return self.definitionToReference().parameterConnectionsTable()
    
    def getMinimalNodes(self):
        return self.definitionToReference().getMinimalNodes()
    
    def getMaximalNodes(self):
        return self.definitionToReference().getMaximalNodes()
    
    def getIdForParameterReference(self, node, parameterId):
        return self.definitionToReference().getIdForParameterReference(node, parameterId)

    def hasParameterSweep(self):
        return len(self.parameterSweeps()) is not 0

    def isParameterSweep(self, parameterId, value=None):
        
        if value is not None:
            if value:
                self.parameterSweeps().add(parameterId)
            else:
                self.parameterSweeps().discard(parameterId)
        
        return parameterId in self.parameterSweeps()
    
    
    def addParameterSweepGroup(self, groupMembers):
        unknownSweeps = [x for x in groupMembers if x not in self.parameterSweeps()]
        if len(unknownSweeps) is not 0:
            raise NotImplementedError('cannot not group for unknown sweeps %s' % unknownSweeps)

        # TODO:
        # verify that the sweeps are not already in another group
        # in which case, we have to add the ability to add to a group
        
        groupMembersKey = tuple(groupMembers)
        for groupMember in groupMembers:
            row = self.parameterSweepGroups().addRow()
            row.setColumn('id', groupMember)
            row.setColumn('group', groupMembersKey)
            pass
        
        return

    def getGroupForParameterSweep(self, parameterId):
        filter = RelationalModule.ColumnValueFilter(
            'id',
            FilterModule.EquivalenceFilter(parameterId)
        )
        groups = RelationalModule.Table.reduceRetrieve(
            self.parameterSweepGroups(),
            filter,
            ['group'],
            []
        )
        # by default, a parameter sweep is in its own group
        if len(groups) is 0:
            return (parameterId)
        elif len(groups) is not 1:
            raise NotImplementedError(
                'have not implemented case where parameter sweep is in 2 different groups')
        return groups[0]
    
    
    def isReadyToExecute(self, parentTask):
        allPredecessors = set(self.predecessors())
        
        # need to filter for the predecessors that have completed
        predecessorFilter = FilterModule.constructOrFilter()
        for predecessor in allPredecessors:
            predecessorFilter.addFilter(
                RelationalModule.ColumnValueFilter(
                    'definition',
                    FilterModule.IdentityFilter(predecessor)
                )
            )
            pass
        
        completedPredecessorFilter = FilterModule.constructAndFilter()
        completedPredecessorFilter.addFilter(predecessorFilter)
        completedPredecessorFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'status',
                FilterModule.EquivalenceFilter('completed')
            )
        )
        completedPredecessors = RelationalModule.Table.reduceRetrieve(
            parentTask.tasksTable(),
            completedPredecessorFilter,
            ['definition'],
            []
        )
        
        # if every one of those predecessors
        # can be found in tokens, and vice versa
        return len(set(completedPredecessors).symmetric_difference(allPredecessors)) is 0
    
    
    def parameterStagingRequired(self, parameter, value=None):

        if value is not None:
            self.parameterStagingMap()[parameter] = value
            
        # default to False
        return self.parameterStagingMap().get(parameter, False)
    
    
    # END class ReferenceDefinition
    pass




def getNewNestDefinition():
    return NestDefinition()




from __future__ import with_statement

import pickle

import pomsets.resource as ResourceModule



def pickleDefinition(path, definition):

    pickleCreated = False
    with open(path, 'w') as f:
        pickle.dump(definition, f)
        pickleCreated = True
        pass

    if not pickleCreated:
        raise IOError('failed on creating pickle')

    return



def savePomset(context):
    savePomsetAs(context, context.url())
    return


def savePomsetAs(context, path, extension='.pomset'):

    pomset = context.pomset()

    if not path.endswith(extension):
        path = '.'.join([path, extension])

    try:
        pickleDefinition(path, pomset)

        # saved, so no longer modified
        context.isModified(False)

        # TODO:
        # this should be a full URL
        context.url(path)
    except IOError, e:
        logging.error(e)
        raise

    return


# TODO:
# the value passed in should be a full URL
def loadPomset(path=None):

    pomset = None
    with open(path, 'r') as f:
        pomset = pickle.load(f)
        pass

    if pomset is None:
        # TODO: should send a failed command event
        raise IOError("failed to load pomset from %s" % path)

    pomsetContext = Context()
    pomsetContext.pomset(pomset)

    pomsetContext.isModified(False)

    # TODO:
    # the value passed in should be a full URL
    pomsetContext.url(path)

    return pomsetContext


class Context(ResourceModule.Struct):

    ATTRIBUTES = [
        'isModified',
        'pomset',
        'url'
        ]

    def __init__(self):
        ResourceModule.Struct.__init__(self)
        return
    
    def getContextPathFor(self, pomset):

        # TODO:
        # remove the hardcoding
        # and actually handle pomset nests

        return [self.pomset()]
    
    # END class Context
    pass
